@@ -1,7 +1,7 @@
 /*
  * sip-connection-manager.c - Source for TpsipConnectionManager
  * Copyright (C) 2005-2007 Collabora Ltd.
- * Copyright (C) 2005-2008 Nokia Corporation
+ * Copyright (C) 2005-2009 Nokia Corporation
  *   @author Kai Vehmanen <first.surname@nokia.com>
  *   @author Martti Mela <first.surname@nokia.com>
  *   @author Mikhail Zabaluev <mikhail.zabaluev@nokia.com>
@@ -24,6 +24,8 @@
  * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
+#include "config.h"
+
 #include <stdlib.h>
 #include <string.h>
 
@@ -35,12 +37,21 @@
 #include <tpsip/sofia-decls.h>
 #include <sofia-sip/su_glib.h>
 
+#ifdef HAVE_LIBIPHB
+#include <libiphb.h>
+#endif
+
 #include "sip-connection-manager.h"
 #include "sip-connection.h"
 
 #define DEBUG_FLAG TPSIP_DEBUG_CONNECTION
 #include "debug.h"
 
+/* Time bounds for heartbeat, in seconds */
+#define TPSIP_HEARTBEAT_MIN 0
+#define TPSIP_HEARTBEAT_MAX 30
+
+
 G_DEFINE_TYPE(TpsipConnectionManager, tpsip_connection_manager,
     TP_TYPE_BASE_CONNECTION_MANAGER)
 
@@ -58,7 +69,7 @@ typedef struct {
     gboolean loose_routing;
     gboolean discover_binding;
     gchar *keepalive_mechanism;
-    gint keepalive_interval;
+    guint keepalive_interval;
     gboolean discover_stun;
     gchar *stun_server;
     guint stun_port;
@@ -162,8 +173,8 @@ static const TpCMParamSpec tpsip_params[] = {
     /* Mechanism used for connection keepalive maintenance */
     { "keepalive-mechanism", DBUS_TYPE_STRING_AS_STRING, G_TYPE_STRING,
       0, NULL, G_STRUCT_OFFSET (TpsipConnParams, keepalive_mechanism) },
-    /* KA interval */
-    { "keepalive-interval", DBUS_TYPE_INT32_AS_STRING, G_TYPE_INT,
+    /* Keep-alive interval */
+    { "keepalive-interval", DBUS_TYPE_UINT32_AS_STRING, G_TYPE_UINT,
       0, NULL, G_STRUCT_OFFSET (TpsipConnParams, keepalive_interval) },
     /* Use SRV DNS lookup to discover STUN server */
     { "discover-stun", DBUS_TYPE_BOOLEAN_AS_STRING, G_TYPE_BOOLEAN,
@@ -199,9 +210,102 @@ const TpCMProtocolSpec tpsip_protocols[] = {
 struct _TpsipConnectionManagerPrivate
 {
   su_root_t *sofia_root;
+
+#ifdef HAVE_LIBIPHB
+  iphb_t    heartbeat;
+  su_wait_t heartbeat_wait[1];
+  int       heartbeat_wait_id;
+#endif
 };
 
-#define TPSIP_CONNECTION_MANAGER_GET_PRIVATE(o)     (G_TYPE_INSTANCE_GET_PRIVATE ((o), TPSIP_TYPE_CONNECTION_MANAGER, TpsipConnectionManagerPrivate))
+#define TPSIP_CONNECTION_MANAGER_GET_PRIVATE(obj) ((obj)->priv)
+
+#ifdef HAVE_LIBIPHB
+
+static void heartbeat_shutdown (TpsipConnectionManager *self);
+
+static int
+heartbeat_wakeup (TpsipConnectionManager *self,
+                  su_wait_t *wait,
+                  void *user_data)
+{
+  TpsipConnectionManagerPrivate *priv = TPSIP_CONNECTION_MANAGER_GET_PRIVATE (self);
+
+  DEBUG("tick");
+
+  g_assert (priv->heartbeat != NULL);
+
+  if ((wait->revents & (SU_WAIT_HUP | SU_WAIT_ERR)) != 0)
+    {
+      g_warning ("heartbeat descriptor invalidated prematurely with event mask %hd", wait->revents);
+      heartbeat_shutdown (self);
+      return 0;
+    }
+
+  iphb_wait (priv->heartbeat, TPSIP_HEARTBEAT_MIN, TPSIP_HEARTBEAT_MAX, 0);
+
+  return 0;
+}
+
+#endif /* HAVE_LIBIPHB */
+
+static void
+heartbeat_init (TpsipConnectionManager *self)
+{
+#ifdef HAVE_LIBIPHB
+  TpsipConnectionManagerPrivate *priv = TPSIP_CONNECTION_MANAGER_GET_PRIVATE (self);
+  int wait_id;
+  int reference_interval = 0;
+
+  su_root_set_max_defer (priv->sofia_root, TPSIP_HEARTBEAT_MAX * 1000L + 50L);
+
+  priv->heartbeat = iphb_open (&reference_interval);
+
+  if (priv->heartbeat == NULL)
+    {
+      g_warning ("opening IP heartbeat failed: %s", strerror (errno));
+      return;
+    }
+
+  DEBUG("heartbeat opened with reference interval %d", reference_interval);
+
+  su_wait_init (priv->heartbeat_wait);
+  if (su_wait_create (priv->heartbeat_wait,
+                      iphb_get_fd (priv->heartbeat),
+                      SU_WAIT_IN) != 0)
+    g_critical ("could not create a wait object");
+
+  wait_id = su_root_register (priv->sofia_root,
+      priv->heartbeat_wait, heartbeat_wakeup, NULL, 0);
+
+  g_return_if_fail (wait_id > 0);
+  priv->heartbeat_wait_id = wait_id;
+
+  /* Prime the heartbeat for the first time.
+   * The correct sequence is iphb_wait() -> poll */
+  iphb_wait (priv->heartbeat, TPSIP_HEARTBEAT_MIN, TPSIP_HEARTBEAT_MAX, 0);
+
+#endif /* HAVE_LIBIPHB */
+}
+
+static void
+heartbeat_shutdown (TpsipConnectionManager *self)
+{
+#ifdef HAVE_LIBIPHB
+  TpsipConnectionManagerPrivate *priv = TPSIP_CONNECTION_MANAGER_GET_PRIVATE (self);
+
+  if (priv->heartbeat_wait_id == 0)
+    return;
+
+  su_root_deregister (priv->sofia_root, priv->heartbeat_wait_id);
+  priv->heartbeat_wait_id = 0;
+
+  su_wait_destroy (priv->heartbeat_wait);
+
+  iphb_close (priv->heartbeat);
+  priv->heartbeat = NULL;
+#endif /* HAVE_LIBIPHB */
+}
 
 static void
 tpsip_connection_manager_init (TpsipConnectionManager *obj)
@@ -216,6 +320,8 @@ tpsip_connection_manager_init (TpsipConnectionManager *obj)
   su_root_threading(priv->sofia_root, 0);
   source = su_root_gsource(priv->sofia_root);
   g_source_attach(source, NULL);
+
+  heartbeat_init (obj);
 }
 
 static void tpsip_connection_manager_finalize (GObject *object);
@@ -246,6 +352,8 @@ tpsip_connection_manager_finalize (GObject *object)
   TpsipConnectionManagerPrivate *priv = TPSIP_CONNECTION_MANAGER_GET_PRIVATE (self);
   GSource *source;
 
+  heartbeat_shutdown (self);
+
   source = su_root_gsource(priv->sofia_root);
   g_source_destroy(source);
   su_root_destroy(priv->sofia_root);
