@@ -36,9 +36,6 @@
 #include <telepathy-glib/interfaces.h>
 #include <telepathy-glib/svc-media-interfaces.h>
 
-/* Hold interface */
-#include <tpsip-extensions/extensions.h>
-
 #include "config.h"
 
 #include "sip-media-session.h"
@@ -50,6 +47,11 @@
 #define DEBUG_FLAG TPSIP_DEBUG_MEDIA
 #include "debug.h"
 
+/* The timeout for outstanding re-INVITE transactions in seconds.
+ * Chosen to match the allowed cancellation timeout for proxies
+ * described in RFC 3261 Section 13.3.1.1 */
+#define TPSIP_REINVITE_TIMEOUT 180
+
 static void session_handler_iface_init (gpointer, gpointer);
 
 G_DEFINE_TYPE_WITH_CODE(TpsipMediaSession,
@@ -59,8 +61,6 @@ G_DEFINE_TYPE_WITH_CODE(TpsipMediaSession,
       session_handler_iface_init)
     )
 
-#define DEFAULT_SESSION_TIMEOUT 50000
-
 /* signal enum */
 enum
 {
@@ -77,6 +77,7 @@ enum
   PROP_PEER,
   PROP_HOLD_STATE,
   PROP_HOLD_STATE_REASON,
+  PROP_RTCP_ENABLED,
   PROP_LOCAL_IP_ADDRESS,
   LAST_PROPERTY
 };
@@ -107,6 +108,7 @@ static const char* session_states[] =
     "active",
     "reinvite-sent",
     "reinvite-received",
+    "reinvite-pending",
     "ended"
 };
 
@@ -122,19 +124,20 @@ struct _TpsipMediaSessionPrivate
   nua_handle_t *nua_op;                   /* see gobj. prop. 'nua-handle' */
   TpHandle peer;                          /* see gobj. prop. 'peer' */
   gchar *local_ip_address;                /* see gobj. prop. 'local-ip-address' */
+  gboolean rtcp_enabled;                  /* see gobj. prop. 'rtcp-enabled' */
   TpsipMediaSessionState state;           /* session state */
-  TpsipLocalHoldState hold_state;         /* local hold state aggregated from stream directions */
-  TpsipLocalHoldStateReason hold_reason;  /* last used hold state change reason */
+  TpLocalHoldState hold_state;         /* local hold state aggregated from stream directions */
+  TpLocalHoldStateReason hold_reason;  /* last used hold state change reason */
   nua_saved_event_t saved_event[1];       /* Saved incoming request event */
   gint local_non_ready;                   /* number of streams with local information update pending */
   guint remote_stream_count;              /* number of streams last seen in a remote offer */
-  guint catcher_id;
-  guint timer_id;
+  guint glare_timer_id;
   su_home_t *home;                        /* Sofia memory home for remote SDP session structure */
   su_home_t *backup_home;                 /* Sofia memory home for previous generation remote SDP session*/
   sdp_session_t *remote_sdp;              /* last received remote session */
   sdp_session_t *backup_remote_sdp;       /* previous remote session */
   GPtrArray *streams;
+  gboolean remote_initiated;              /*< session is remotely intiated */
   gboolean accepted;                      /*< session has been locally accepted for use */
   gboolean se_ready;                      /*< connection established with stream-engine */
   gboolean pending_offer;                 /*< local media have been changed, but a re-INVITE is pending */
@@ -157,8 +160,6 @@ tpsip_media_session_get_stream (TpsipMediaSession *self,
                               guint stream_id,
                               GError **error);
 
-static gboolean priv_catch_remote_nonupdate (gpointer data);
-static gboolean priv_timeout_session (gpointer data);
 static TpsipMediaStream* priv_create_media_stream (TpsipMediaSession *session,
                                                  guint media_type,
                                                  guint pending_send_flags);
@@ -175,8 +176,9 @@ static void tpsip_media_session_init (TpsipMediaSession *obj)
   TpsipMediaSessionPrivate *priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (obj);
 
   priv->state = TPSIP_MEDIA_SESSION_STATE_CREATED;
-  priv->hold_state = TPSIP_LOCAL_HOLD_STATE_UNHELD;
-  priv->hold_reason = TPSIP_LOCAL_HOLD_STATE_REASON_NONE;
+  priv->hold_state = TP_LOCAL_HOLD_STATE_UNHELD;
+  priv->hold_reason = TP_LOCAL_HOLD_STATE_REASON_NONE;
+  priv->rtcp_enabled = TRUE;
 
   /* allocate any data required by the object here */
   priv->streams = g_ptr_array_new ();
@@ -230,6 +232,9 @@ static void tpsip_media_session_get_property (GObject    *object,
     case PROP_LOCAL_IP_ADDRESS:
       g_value_set_string (value, priv->local_ip_address);
       break;
+    case PROP_RTCP_ENABLED:
+      g_value_set_boolean (value, priv->rtcp_enabled);
+      break;
     default:
       G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
       break;
@@ -292,79 +297,58 @@ tpsip_media_session_class_init (TpsipMediaSessionClass *klass)
   object_class->finalize = tpsip_media_session_finalize;
 
   param_spec = g_param_spec_object ("media-channel", "TpsipMediaChannel object",
-                                    "SIP media channel object that owns this "
-                                    "media session object (not reference counted).",
-                                    TPSIP_TYPE_MEDIA_CHANNEL,
-                                    G_PARAM_CONSTRUCT_ONLY |
-                                    G_PARAM_READWRITE |
-                                    G_PARAM_STATIC_NICK |
-                                    G_PARAM_STATIC_BLURB);
+      "SIP media channel object that owns this media session object"
+      " (not reference counted).",
+      TPSIP_TYPE_MEDIA_CHANNEL,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_MEDIA_CHANNEL, param_spec);
 
   param_spec = g_param_spec_string ("object-path", "D-Bus object path",
-                                    "The D-Bus object path used for this "
-                                    "object on the bus.",
-                                    NULL,
-                                    G_PARAM_CONSTRUCT_ONLY |
-                                    G_PARAM_READWRITE |
-                                    G_PARAM_STATIC_NAME |
-                                    G_PARAM_STATIC_BLURB);
+      "The D-Bus object path used for this object on the bus.",
+      NULL,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_OBJECT_PATH, param_spec);
 
-  param_spec = g_param_spec_pointer("nua-handle", "Sofia-SIP NUA operator handle",
-                                    "NUA stack operation handle associated "
-                                    "with this media session.",
-                                    G_PARAM_CONSTRUCT_ONLY |
-                                    G_PARAM_READWRITE |
-                                    G_PARAM_STATIC_NAME |
-                                    G_PARAM_STATIC_BLURB);
+  param_spec = g_param_spec_pointer ("nua-handle", "NUA handle",
+      "NUA stack operation handle associated with this media session.",
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_NUA_OP, param_spec);
 
   param_spec = g_param_spec_uint ("peer", "Session peer",
-                                  "The TpHandle representing the contact "
-                                  "with whom this session communicates.",
-                                  0, G_MAXUINT32, 0,
-                                  G_PARAM_CONSTRUCT_ONLY |
-                                  G_PARAM_READWRITE |
-                                  G_PARAM_STATIC_NAME |
-                                  G_PARAM_STATIC_BLURB);
+      "The TpHandle representing the contact with whom this session communicates.",
+      0, G_MAXUINT32,
+      0,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_PEER, param_spec);
 
   param_spec = g_param_spec_uint ("hold-state", "Local hold state",
-                                  "The current Local_Hold_State value"
-                                  " as reported by the Hold interface"
-                                  " for the channel",
-                                  TPSIP_LOCAL_HOLD_STATE_UNHELD,
-                                  TPSIP_LOCAL_HOLD_STATE_PENDING_UNHOLD,
-                                  TPSIP_LOCAL_HOLD_STATE_UNHELD,
-                                  G_PARAM_READABLE |
-                                  G_PARAM_STATIC_NAME |
-                                  G_PARAM_STATIC_BLURB);
+      "The current Local_Hold_State value as reported by the Hold interface",
+      TP_LOCAL_HOLD_STATE_UNHELD, TP_LOCAL_HOLD_STATE_PENDING_UNHOLD,
+      TP_LOCAL_HOLD_STATE_UNHELD,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_HOLD_STATE, param_spec);
 
   param_spec = g_param_spec_uint ("hold-state-reason",
-                                  "Local hold state change reason",
-                                  "The last Local_Hold_State_Reason value"
-                                  " as reported by the Hold interface"
-                                  " for the channel",
-                                  TPSIP_LOCAL_HOLD_STATE_REASON_NONE,
-                                  TPSIP_LOCAL_HOLD_STATE_REASON_RESOURCE_NOT_AVAILABLE,
-                                  TPSIP_LOCAL_HOLD_STATE_REASON_NONE,
-                                  G_PARAM_READABLE |
-                                  G_PARAM_STATIC_NAME |
-                                  G_PARAM_STATIC_BLURB);
+      "Local hold state change reason",
+      "The last Local_Hold_State_Reason value as reported by the Hold interface",
+      TP_LOCAL_HOLD_STATE_REASON_NONE,
+      TP_LOCAL_HOLD_STATE_REASON_RESOURCE_NOT_AVAILABLE,
+      TP_LOCAL_HOLD_STATE_REASON_NONE,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_HOLD_STATE_REASON, param_spec);
 
   param_spec = g_param_spec_string ("local-ip-address", "Local IP address",
-                                    "The local IP address preferred for "
-                                    "media streams",
-                                    NULL,
-                                    G_PARAM_CONSTRUCT_ONLY |
-                                    G_PARAM_READWRITE |
-                                    G_PARAM_STATIC_NAME |
-                                    G_PARAM_STATIC_BLURB);
+      "The local IP address preferred for media streams",
+      NULL,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_LOCAL_IP_ADDRESS, param_spec);
 
+  param_spec = g_param_spec_boolean ("rtcp-enabled", "RTCP enabled",
+      "Is RTCP enabled session-wide",
+      TRUE,
+      G_PARAM_READABLE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_RTCP_ENABLED, param_spec);
+
   signals[SIG_STATE_CHANGED] =
     g_signal_new ("state-changed",
                   G_OBJECT_CLASS_TYPE (klass),
@@ -388,11 +372,8 @@ tpsip_media_session_dispose (GObject *object)
 
   priv->dispose_has_run = TRUE;
 
-  if (priv->catcher_id)
-    g_source_remove (priv->catcher_id);
-
-  if (priv->timer_id)
-    g_source_remove (priv->timer_id);
+  if (priv->glare_timer_id)
+    g_source_remove (priv->glare_timer_id);
 
   if (G_OBJECT_CLASS (tpsip_media_session_parent_class)->dispose)
     G_OBJECT_CLASS (tpsip_media_session_parent_class)->dispose (object);
@@ -453,7 +434,7 @@ tpsip_media_session_error (TpSvcMediaSessionHandler *iface,
 {
   TpsipMediaSession *obj = TPSIP_MEDIA_SESSION (iface);
 
-  SESSION_DEBUG(obj, "Media.SessionHandler::Error called (%s) terminating session", message);
+  SESSION_DEBUG (obj, "Media.SessionHandler::Error called (%s), terminating session", message);
 
   tpsip_media_session_terminate (obj);
 
@@ -574,6 +555,11 @@ priv_apply_streams_pending_send (TpsipMediaSession *session,
   TpsipMediaStream *stream;
   guint i;
 
+  /* If there has been a local change pending a re-INVITE,
+   * leave pending remote send for the next transaction */
+  if (priv->pending_offer)
+    pending_send_mask &= ~(guint)TP_MEDIA_STREAM_PENDING_REMOTE_SEND;
+
   /* Apply the local pending send flags where applicable */
   for (i = 0; i < priv->streams->len; i++)
     {
@@ -596,53 +582,39 @@ tpsip_media_session_change_state (TpsipMediaSession *session,
   old_state = priv->state;
   priv->state = new_state;
 
-  SESSION_DEBUG(session, "state change: %s -> %s",
-                session_states[old_state],
-                session_states[new_state]);
+  SESSION_DEBUG (session, "state change: %s -> %s",
+      session_states[old_state],
+      session_states[new_state]);
 
   switch (new_state)
     {
     case TPSIP_MEDIA_SESSION_STATE_CREATED:
-      break;
-    case TPSIP_MEDIA_SESSION_STATE_ENDED:
-      priv_close_all_streams (session);
-      DEBUG("releasing the NUA handle %p", priv->nua_op);
-      if (priv->nua_op != NULL)
-        {
-          nua_handle_unref (priv->nua_op);
-          priv->nua_op = NULL;
-        }
-      break;
     case TPSIP_MEDIA_SESSION_STATE_INVITE_RECEIVED:
     case TPSIP_MEDIA_SESSION_STATE_REINVITE_RECEIVED:
-      priv->catcher_id = g_idle_add (priv_catch_remote_nonupdate, session);
-      /* Fall through to the next case */
     case TPSIP_MEDIA_SESSION_STATE_INVITE_SENT:
     case TPSIP_MEDIA_SESSION_STATE_REINVITE_SENT:
-      if (priv->timer_id)
-        {
-          g_source_remove (priv->timer_id);
-        }
-      priv->timer_id =
-        g_timeout_add (DEFAULT_SESSION_TIMEOUT, priv_timeout_session, session);
-      break;
     case TPSIP_MEDIA_SESSION_STATE_RESPONSE_RECEIVED:
+    case TPSIP_MEDIA_SESSION_STATE_REINVITE_PENDING:
       break;
     case TPSIP_MEDIA_SESSION_STATE_ACTIVE:
-      if (priv->timer_id)
-        {
-	  g_source_remove (priv->timer_id);
-	  priv->timer_id = 0;
-        }
       /* Apply any pending remote send after outgoing INVITEs.
        * We don't want automatic removal of pending local send after
        * responding to incoming re-INVITEs, however */
       priv_apply_streams_pending_send (session,
                                        TP_MEDIA_STREAM_PENDING_REMOTE_SEND);
       break;
+    case TPSIP_MEDIA_SESSION_STATE_ENDED:
+      priv_close_all_streams (session);
+      DEBUG("destroying the NUA handle %p", priv->nua_op);
+      if (priv->nua_op != NULL)
+        {
+          nua_handle_destroy (priv->nua_op);
+          priv->nua_op = NULL;
+        }
+      break;
 
-    /* Don't add default because we want to be warned by the compiler
-     * about unhandled states */
+      /* Don't add default because we want to be warned by the compiler
+       * about unhandled states */
     }
 
   g_signal_emit (session, signals[SIG_STATE_CHANGED], 0, old_state, new_state);
@@ -654,13 +626,14 @@ tpsip_media_session_change_state (TpsipMediaSession *session,
 #ifdef ENABLE_DEBUG
 void
 tpsip_media_session_debug (TpsipMediaSession *session,
-			 const gchar *format, ...)
+                           const gchar *format, ...)
 {
-  va_list list;
-  gchar buf[512];
   TpsipMediaSessionPrivate *priv;
+  va_list list;
+  gchar buf[240];
 
-  g_assert (TPSIP_IS_MEDIA_SESSION (session));
+  if (!tpsip_debug_flag_is_set (DEBUG_FLAG))
+    return;
 
   priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (session);
 
@@ -676,65 +649,6 @@ tpsip_media_session_debug (TpsipMediaSession *session,
 }
 #endif /* ENABLE_DEBUG */
 
-static gboolean
-priv_catch_remote_nonupdate (gpointer data)
-{
-  TpsipMediaSession *session = data;
-
-  DEBUG("called");
-
-  /* Accordingly to the last experimental data, non-modifying INVITEs
-   * cause the stack to emit nua_i_state nonetheless */
-  g_assert_not_reached();
-
-  /* TODO: figure out what happens in the 3pcc scenario when we get
-   * an INVITE but no session offer */
-
-  /* Should do the right thing if there were no remote media updates */
-  priv_request_response_step (session);
-
-  return FALSE;
-}
-
-static gboolean priv_timeout_session (gpointer data)
-{
-  TpsipMediaSession *session = data;
-  TpsipMediaSessionPrivate *priv;
-  TpChannelGroupChangeReason reason;
-  gboolean change = FALSE;
-  TpHandle actor;
-
-  DEBUG("session timed out");
-
-  priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (session); 
-
-  if (priv->state == TPSIP_MEDIA_SESSION_STATE_INVITE_SENT)
-    {
-      reason = TP_CHANNEL_GROUP_CHANGE_REASON_NO_ANSWER;
-      actor = 0;
-      change = TRUE;
-    }
-  else if (priv->state == TPSIP_MEDIA_SESSION_STATE_INVITE_RECEIVED)
-    {
-      reason = TP_CHANNEL_GROUP_CHANGE_REASON_NONE;
-      actor = priv->peer;
-      change = TRUE;
-    }
-
-  if (change)
-    {
-      TpIntSet *set = tp_intset_new ();
-      tp_intset_add (set, priv->peer);
-      tp_group_mixin_change_members (G_OBJECT (priv->channel), "Timed out",
-                                     NULL, set, NULL, NULL, actor, reason);
-      tp_intset_destroy (set);
-    }
-
-  tpsip_media_session_terminate (session);
-
-  return FALSE;
-}
-
 void tpsip_media_session_terminate (TpsipMediaSession *session)
 {
   TpsipMediaSessionPrivate *priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (session);
@@ -749,11 +663,14 @@ void tpsip_media_session_terminate (TpsipMediaSession *session)
 
   if (priv->nua_op != NULL)
     {
+      /* XXX: should the stack do pretty much the same
+       * (except freeing the saved event) upon nua_handle_destroy()? */
       switch (priv->state)
         {
         case TPSIP_MEDIA_SESSION_STATE_ACTIVE:
         case TPSIP_MEDIA_SESSION_STATE_RESPONSE_RECEIVED:
         case TPSIP_MEDIA_SESSION_STATE_REINVITE_SENT:
+        case TPSIP_MEDIA_SESSION_STATE_REINVITE_PENDING:
           DEBUG("sending BYE");
           nua_bye (priv->nua_op, TAG_END());
           break;
@@ -790,16 +707,10 @@ tpsip_media_session_set_remote_media (TpsipMediaSession *session,
                                     const sdp_session_t* sdp)
 {
   TpsipMediaSessionPrivate *priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (session);
+  gboolean authoritative;
 
   DEBUG ("enter");
 
-  /* Remove the non-update catcher because we've got an update */
-  if (priv->catcher_id)
-    {
-      g_source_remove (priv->catcher_id);
-      priv->catcher_id = 0;
-    }
-
   if (priv->state == TPSIP_MEDIA_SESSION_STATE_INVITE_SENT
       || priv->state == TPSIP_MEDIA_SESSION_STATE_REINVITE_SENT)
     {
@@ -820,13 +731,9 @@ tpsip_media_session_set_remote_media (TpsipMediaSession *session,
       priv->remote_stream_count = count;
     }
 
-  /* Handle session non-updates */
+  /* Shortcut session non-updates */
   if (!sdp_session_cmp (priv->remote_sdp, sdp))
-    {
-      /* Should do the proper response etc. */
-      priv_request_response_step (session);
-      return TRUE;
-    }
+    goto finally;
 
   /* Delete a backup session structure, if any */
   if (priv->backup_remote_sdp != NULL)
@@ -851,10 +758,16 @@ tpsip_media_session_set_remote_media (TpsipMediaSession *session,
   priv->remote_sdp = sdp_session_dup (priv->home, sdp);
   g_return_val_if_fail (priv->remote_sdp != NULL, FALSE);
 
-  return priv_update_remote_media (
-                session,
-                (priv->state == TPSIP_MEDIA_SESSION_STATE_INVITE_RECEIVED
-                 || priv->state == TPSIP_MEDIA_SESSION_STATE_REINVITE_RECEIVED));
+  authoritative = (priv->state == TPSIP_MEDIA_SESSION_STATE_INVITE_RECEIVED
+                || priv->state == TPSIP_MEDIA_SESSION_STATE_REINVITE_RECEIVED);
+  if (!priv_update_remote_media (session, authoritative))
+    return FALSE;
+
+finally:
+  /* Make sure to always transition states and send out the response,
+   * even if no stream-engine roundtrips were initiated */
+  priv_request_response_step (session);
+  return TRUE;
 }
 
 void
@@ -939,6 +852,8 @@ gboolean tpsip_media_session_request_streams (TpsipMediaSession *session,
     priv_add_stream_list_entry (ret, stream, session);
   }
 
+  priv_local_media_changed (session);
+
   return TRUE;
 }
 
@@ -999,7 +914,8 @@ tpsip_media_session_request_stream_direction (TpsipMediaSession *self,
       return FALSE;
     }
 
-  SESSION_DEBUG (self, "direction %u requested for stream %u", direction, stream_id);
+  SESSION_DEBUG (self, "direction %u requested for stream %u",
+      direction, stream_id);
 
   if (priv->state == TPSIP_MEDIA_SESSION_STATE_INVITE_RECEIVED
       || priv->state == TPSIP_MEDIA_SESSION_STATE_REINVITE_RECEIVED)
@@ -1063,6 +979,8 @@ tpsip_media_session_receive_invite (TpsipMediaSession *self)
   g_return_if_fail (priv->state == TPSIP_MEDIA_SESSION_STATE_CREATED);  
   g_return_if_fail (priv->nua_op != NULL);
 
+  priv->remote_initiated = TRUE;
+
   nua_respond (priv->nua_op, SIP_180_RINGING, TAG_END());
 
   tpsip_media_session_change_state (self, TPSIP_MEDIA_SESSION_STATE_INVITE_RECEIVED);
@@ -1073,8 +991,18 @@ tpsip_media_session_receive_reinvite (TpsipMediaSession *self)
 {
   TpsipMediaSessionPrivate *priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (self);
 
-  g_return_if_fail (priv->state == TPSIP_MEDIA_SESSION_STATE_ACTIVE
-                    || priv->state == TPSIP_MEDIA_SESSION_STATE_RESPONSE_RECEIVED);
+  /* Check for permitted state transitions */
+  switch (priv->state)
+    {
+    case TPSIP_MEDIA_SESSION_STATE_ACTIVE:
+    case TPSIP_MEDIA_SESSION_STATE_RESPONSE_RECEIVED:
+      break;
+    case TPSIP_MEDIA_SESSION_STATE_REINVITE_PENDING:
+      g_source_remove (priv->glare_timer_id);
+      break;
+    default:
+      g_return_if_reached ();
+    }
 
   priv_save_event (self);
 
@@ -1089,7 +1017,7 @@ tpsip_media_session_accept (TpsipMediaSession *self)
   if (priv->accepted)
     return;
 
-  SESSION_DEBUG(self, "accepting the session");
+  SESSION_DEBUG (self, "accepting the session");
 
   priv->accepted = TRUE;
 
@@ -1103,17 +1031,17 @@ tpsip_media_session_accept (TpsipMediaSession *self)
 }
 
 void
-tpsip_media_session_reject (TpsipMediaSession *self,
-                          gint status,
-                          const char *message)
+tpsip_media_session_respond (TpsipMediaSession *self,
+                             gint status,
+                             const char *message)
 {
   TpsipMediaSessionPrivate *priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (self);
 
+  SESSION_DEBUG (self, "responding: %03d %s", status, message ? message : "");
+
   if (message != NULL && !message[0])
     message = NULL;
 
-  SESSION_DEBUG(self, "responding: %03d %s", status, message == NULL? "" : message);
-
   if (priv->nua_op)
     nua_respond (priv->nua_op, status, message, TAG_END());
 }
@@ -1124,6 +1052,62 @@ gboolean tpsip_media_session_is_accepted (TpsipMediaSession *self)
   return priv->accepted;
 }
 
+static gboolean
+priv_glare_retry (gpointer session)
+{
+  TpsipMediaSession *self = session;
+  TpsipMediaSessionPrivate *priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (self);
+
+  SESSION_DEBUG (self, "glare resolution interval is over");
+
+  if (priv->state == TPSIP_MEDIA_SESSION_STATE_REINVITE_PENDING)
+    priv_session_invite (self, TRUE);
+
+  /* Reap the timer */
+  priv->glare_timer_id = 0;
+  return FALSE;
+}
+
+void
+tpsip_media_session_resolve_glare (TpsipMediaSession *self)
+{
+  TpsipMediaSessionPrivate *priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (self);
+  guint interval;
+
+  if (priv->state != TPSIP_MEDIA_SESSION_STATE_REINVITE_SENT)
+    {
+      SESSION_DEBUG (self, "glare resolution triggered in unexpected state");
+      return;
+    }
+
+  /*
+   * Set the grace interval accordinlgly to RFC 3261 section 14.1:
+   *
+   *  1. If the UAC is the owner of the Call-ID of the dialog ID
+   *     (meaning it generated the value), T has a randomly chosen value
+   *     between 2.1 and 4 seconds in units of 10 ms.
+   *  2. If the UAC is not the owner of the Call-ID of the dialog ID, T
+   *     has a randomly chosen value of between 0 and 2 seconds in units
+   *     of 10 ms.
+   */
+  if (priv->pending_offer)
+    interval = 0;       /* cut short, we have new things to negotiate */
+  else if (priv->remote_initiated)
+    interval = g_random_int_range (0, 200) * 10;
+  else
+    interval = g_random_int_range (210, 400) * 10;
+
+  if (priv->glare_timer_id != 0)
+    g_source_remove (priv->glare_timer_id);
+
+  priv->glare_timer_id = g_timeout_add (interval, priv_glare_retry, self);
+
+  SESSION_DEBUG (self, "glare resolution interval %u msec", interval);
+
+  tpsip_media_session_change_state (
+        self, TPSIP_MEDIA_SESSION_STATE_REINVITE_PENDING);
+}
+
 static TpsipMediaStream *
 tpsip_media_session_get_stream (TpsipMediaSession *self,
                                 guint stream_id,
@@ -1153,7 +1137,7 @@ tpsip_media_session_get_stream (TpsipMediaSession *self,
   return stream;
 }
 
-TpsipLocalHoldState
+TpLocalHoldState
 tpsip_media_session_get_hold_state (TpsipMediaSession *self)
 {
   TpsipMediaSessionPrivate *priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (self);
@@ -1164,23 +1148,25 @@ static gboolean
 tpsip_media_session_is_local_hold_ongoing (TpsipMediaSession *self)
 {
   TpsipMediaSessionPrivate *priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (self);
-  return (priv->hold_state == TPSIP_LOCAL_HOLD_STATE_HELD
-          || priv->hold_state == TPSIP_LOCAL_HOLD_STATE_PENDING_HOLD);
+  return (priv->hold_state == TP_LOCAL_HOLD_STATE_HELD
+          || priv->hold_state == TP_LOCAL_HOLD_STATE_PENDING_HOLD);
 }
 
 static void
 priv_initiate_hold (TpsipMediaSession *self,
                     gboolean hold,
-                    TpsipLocalHoldStateReason reason)
+                    TpLocalHoldStateReason reason)
 {
   TpsipMediaSessionPrivate *priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (self);
   gboolean stream_hold_requested = FALSE;
   TpsipMediaStream *stream;
   guint i;
 
+  DEBUG("enter");
+
   if (hold)
     {
-      if (priv->hold_state == TPSIP_LOCAL_HOLD_STATE_HELD)
+      if (priv->hold_state == TP_LOCAL_HOLD_STATE_HELD)
         {
           g_message ("redundant hold request");
           return;
@@ -1188,7 +1174,7 @@ priv_initiate_hold (TpsipMediaSession *self,
     }
   else
     {
-      if (priv->hold_state == TPSIP_LOCAL_HOLD_STATE_UNHELD)
+      if (priv->hold_state == TP_LOCAL_HOLD_STATE_UNHELD)
         {
           g_message ("redundant unhold request");
           return;
@@ -1215,20 +1201,19 @@ priv_initiate_hold (TpsipMediaSession *self,
 
   if (stream_hold_requested)
     {
-      priv->hold_state = hold? TPSIP_LOCAL_HOLD_STATE_PENDING_HOLD
-                             : TPSIP_LOCAL_HOLD_STATE_PENDING_UNHOLD;
+      priv->hold_state = hold? TP_LOCAL_HOLD_STATE_PENDING_HOLD
+                             : TP_LOCAL_HOLD_STATE_PENDING_UNHOLD;
     }
   else
     {
       /* There were no streams to flip, short cut to the final state */
-      priv->hold_state = hold? TPSIP_LOCAL_HOLD_STATE_HELD
-                             : TPSIP_LOCAL_HOLD_STATE_UNHELD;
+      priv->hold_state = hold? TP_LOCAL_HOLD_STATE_HELD
+                             : TP_LOCAL_HOLD_STATE_UNHELD;
     }
   priv->hold_reason = reason;
 
-  tpsip_svc_channel_interface_hold_emit_hold_state_changed (priv->channel,
-                                                            priv->hold_state,
-                                                            reason);
+  tp_svc_channel_interface_hold_emit_hold_state_changed (priv->channel,
+      priv->hold_state, reason);
 }
 
 static void
@@ -1236,20 +1221,22 @@ priv_finalize_hold (TpsipMediaSession *self)
 {
   TpsipMediaSessionPrivate *priv = TPSIP_MEDIA_SESSION_GET_PRIVATE (self);
   TpsipMediaStream *stream;
-  TpsipLocalHoldState final_hold_state;
+  TpLocalHoldState final_hold_state;
   guint hold_mask;
   guint unhold_mask;
   guint i;
 
+  DEBUG("enter");
+
   switch (priv->hold_state)
     {
-    case TPSIP_LOCAL_HOLD_STATE_PENDING_HOLD:
-      final_hold_state = TPSIP_LOCAL_HOLD_STATE_HELD;
+    case TP_LOCAL_HOLD_STATE_PENDING_HOLD:
+      final_hold_state = TP_LOCAL_HOLD_STATE_HELD;
       hold_mask = TP_MEDIA_STREAM_DIRECTION_SEND;
       unhold_mask = 0;
       break;
-    case TPSIP_LOCAL_HOLD_STATE_PENDING_UNHOLD:
-      final_hold_state = TPSIP_LOCAL_HOLD_STATE_UNHELD;
+    case TP_LOCAL_HOLD_STATE_PENDING_UNHOLD:
+      final_hold_state = TP_LOCAL_HOLD_STATE_UNHELD;
       hold_mask = TP_MEDIA_STREAM_DIRECTION_BIDIRECTIONAL;
       unhold_mask = TP_MEDIA_STREAM_DIRECTION_RECEIVE;
       break;
@@ -1260,9 +1247,8 @@ priv_finalize_hold (TpsipMediaSession *self)
     }
 
   priv->hold_state = final_hold_state;
-  tpsip_svc_channel_interface_hold_emit_hold_state_changed (priv->channel,
-                                                            final_hold_state,
-                                                            priv->hold_reason);
+  tp_svc_channel_interface_hold_emit_hold_state_changed (priv->channel,
+      final_hold_state, priv->hold_reason);
 
   /* Set stream directions accordingly to the achieved hold state */
   for (i = 0; i < priv->streams->len; i++)
@@ -1286,7 +1272,7 @@ tpsip_media_session_request_hold (TpsipMediaSession *self,
 {
   priv_initiate_hold (self,
                       hold,
-                      TPSIP_LOCAL_HOLD_STATE_REASON_REQUESTED);
+                      TP_LOCAL_HOLD_STATE_REASON_REQUESTED);
 }
 
 gboolean
@@ -1411,6 +1397,7 @@ priv_local_media_changed (TpsipMediaSession *session)
       priv->pending_offer = TRUE;
       break;
     case TPSIP_MEDIA_SESSION_STATE_ACTIVE:
+    case TPSIP_MEDIA_SESSION_STATE_REINVITE_PENDING:
       if (priv->local_non_ready == 0)
         priv_session_invite (session, TRUE);
       else
@@ -1449,6 +1436,8 @@ priv_update_remote_hold (TpsipMediaSession *session)
   if (!has_streams)
     return;
 
+  DEBUG("the session is remotely %s", remote_held? "held" : "unheld");
+
   if (remote_held)
     tpsip_media_channel_change_call_state (priv->channel,
                                            priv->peer,
@@ -1468,10 +1457,16 @@ priv_update_remote_media (TpsipMediaSession *session, gboolean authoritative)
   const sdp_media_t *media;
   gboolean has_supported_media = FALSE;
   guint direction_up_mask;
+  guint pending_send_mask;
   guint i;
 
   g_return_val_if_fail (priv->remote_sdp != NULL, FALSE);
 
+  /* Update the session-wide RTCP enable flag
+   * before updating stream media */
+  priv->rtcp_enabled = !tpsip_sdp_rtcp_bandwidth_throttled (
+                                priv->remote_sdp->sdp_bandwidths);
+
   /*
    * Do not allow:
    * 1) an answer to bump up directions beyond what's been offered;
@@ -1485,13 +1480,21 @@ priv_update_remote_media (TpsipMediaSession *session, gboolean authoritative)
   else
     direction_up_mask = 0;
 
+  /* A remote media requesting to enable sending would need local approval.
+   * Also, if there have been any local media updates pending a re-INVITE,
+   * keep or bump the pending remote send flag on the streams: it will
+   * be resolved in the next re-INVITE transaction */
+  pending_send_mask = TP_MEDIA_STREAM_PENDING_LOCAL_SEND;
+  if (priv->pending_offer)
+    pending_send_mask |= TP_MEDIA_STREAM_PENDING_REMOTE_SEND;
+
   media = priv->remote_sdp->sdp_media;
 
   /* note: for each session, we maintain an ordered list of
    *       streams (SDP m-lines) which are matched 1:1 to
    *       the streams of the remote SDP */
 
-  for (i = 0; media; media = media->m_next, i++)
+  for (i = 0; media != NULL; media = media->m_next, i++)
     {
       TpsipMediaStream *stream = NULL;
       guint media_type;
@@ -1521,15 +1524,13 @@ priv_update_remote_media (TpsipMediaSession *session, gboolean authoritative)
           /* XXX: close this stream and create a new one in its place? */
           g_warning ("The peer has changed the media type, don't know what to do");
         }
-      else
+      else if (tpsip_media_stream_set_remote_media (stream,
+                                                    media,
+                                                    direction_up_mask,
+                                                    pending_send_mask))
         {
-          if (tpsip_media_stream_set_remote_media (stream,
-                                                 media,
-                                                 direction_up_mask))
-            {
-              has_supported_media = TRUE;
-              continue;
-            }
+          has_supported_media = TRUE;
+          continue;
         }
 
       /* There have been problems with the stream update, kill the stream */
@@ -1654,8 +1655,6 @@ priv_session_local_sdp (TpsipMediaSession *session,
         }
     }
 
-  SESSION_DEBUG(session, "generated SDP: {\n%s}", user_sdp->str);
-
   return has_supported_media;
 }
 
@@ -1682,6 +1681,8 @@ priv_session_invite (TpsipMediaSession *session, gboolean reinvite)
                   SOATAG_RTP_SORT(SOA_RTP_SORT_REMOTE),
                   SOATAG_RTP_SELECT(SOA_RTP_SELECT_ALL),
                   NUTAG_AUTOANSWER(0),
+                  TAG_IF(reinvite,
+                         NUTAG_INVITE_TIMER (TPSIP_REINVITE_TIMEOUT)),
                   TAG_END());
       priv->pending_offer = FALSE;
 
@@ -1801,11 +1802,12 @@ priv_request_response_step (TpsipMediaSession *session)
         priv_session_respond (session);
       break;
     case TPSIP_MEDIA_SESSION_STATE_ACTIVE:
+    case TPSIP_MEDIA_SESSION_STATE_REINVITE_PENDING:
       if (priv->pending_offer)
         priv_session_invite (session, TRUE);
       break;
     default:
-      SESSION_DEBUG(session, "no action taken in the current state");
+      SESSION_DEBUG (session, "no action taken in the current state");
     }
 }
 
@@ -1928,10 +1930,10 @@ priv_stream_hold_state_cb (TpsipMediaStream *stream,
   /* Determine the hold state all streams shall come to */
   switch (priv->hold_state)
     {
-    case TPSIP_LOCAL_HOLD_STATE_PENDING_HOLD:
+    case TP_LOCAL_HOLD_STATE_PENDING_HOLD:
       hold = TRUE;
       break;
-    case TPSIP_LOCAL_HOLD_STATE_PENDING_UNHOLD:
+    case TP_LOCAL_HOLD_STATE_PENDING_UNHOLD:
       hold = FALSE;
       break;
     default:
@@ -1939,7 +1941,7 @@ priv_stream_hold_state_cb (TpsipMediaStream *stream,
 
       /* Try to follow the changes and report the resulting hold state */
       g_object_get (stream, "hold-state", &hold, NULL);
-      priv->hold_reason = TPSIP_LOCAL_HOLD_STATE_REASON_NONE;
+      priv->hold_reason = TP_LOCAL_HOLD_STATE_REASON_NONE;
     }
 
   /* Check if all streams have reached the desired hold state */
@@ -1967,7 +1969,7 @@ priv_stream_unhold_failure_cb (TpsipMediaStream *stream,
 {
   priv_initiate_hold (session,
                       TRUE,
-                      TPSIP_LOCAL_HOLD_STATE_REASON_RESOURCE_NOT_AVAILABLE);
+                      TP_LOCAL_HOLD_STATE_REASON_RESOURCE_NOT_AVAILABLE);
 }
 
 static TpsipMediaStream*
@@ -2074,3 +2076,27 @@ session_handler_iface_init (gpointer g_iface, gpointer iface_data)
   IMPLEMENT(ready);
 #undef IMPLEMENT
 }
+
+/* Checks if RTCP is not disabled with bandwidth modifiers
+ * as described in RFC 3556 */
+gboolean
+tpsip_sdp_rtcp_bandwidth_throttled (const sdp_bandwidth_t *b)
+{
+  const sdp_bandwidth_t *b_RS = NULL;
+  const sdp_bandwidth_t *b_RR = NULL;
+
+  while (b != NULL)
+    {
+      if (b->b_modifier_name != NULL)
+        {
+          if (strcmp (b->b_modifier_name, "RS") == 0)
+            b_RS = b;
+          else if (strcmp (b->b_modifier_name, "RR") == 0)
+            b_RR = b;
+        }
+      b = b->b_next;
+    }
+
+  return (b_RS != NULL && b_RS->b_value == 0
+       && b_RR != NULL && b_RR->b_value == 0);
+}
