@@ -73,6 +73,8 @@ enum
 
 static guint signals[SIG_LAST_SIGNAL] = {0};
 
+static GRegex *fmtp_attr_regex = NULL;
+
 /* properties */
 enum
 {
@@ -300,98 +302,62 @@ tpsip_media_stream_class_init (TpsipMediaStreamClass *klass)
   object_class->finalize = tpsip_media_stream_finalize;
 
   param_spec = g_param_spec_object ("media-session", "GabbleMediaSession object",
-                                    "SIP media session object that owns this "
-                                    "media stream object.",
-                                    TPSIP_TYPE_MEDIA_SESSION,
-                                    G_PARAM_CONSTRUCT_ONLY |
-                                    G_PARAM_READWRITE |
-                                    G_PARAM_STATIC_NICK |
-                                    G_PARAM_STATIC_BLURB);
+      "SIP media session object that owns this media stream object.",
+      TPSIP_TYPE_MEDIA_SESSION,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_MEDIA_SESSION, param_spec);
 
   param_spec = g_param_spec_string ("object-path", "D-Bus object path",
-                                    "The D-Bus object path used for this "
-                                    "object on the bus.",
-                                    NULL,
-                                    G_PARAM_CONSTRUCT_ONLY |
-                                    G_PARAM_READWRITE |
-                                    G_PARAM_STATIC_NAME |
-                                    G_PARAM_STATIC_BLURB);
+      "The D-Bus object path used for this object on the bus.",
+      NULL,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_OBJECT_PATH, param_spec);
 
   param_spec = g_param_spec_uint ("id", "Stream ID",
-                                  "A stream number for the stream used in the "
-                                  "D-Bus API.",
-                                  0, G_MAXUINT, 0,
-                                  G_PARAM_CONSTRUCT_ONLY |
-                                  G_PARAM_READWRITE |
-                                  G_PARAM_STATIC_NAME |
-                                  G_PARAM_STATIC_BLURB);
+      "A stream number for the stream used in the D-Bus API.",
+      0, G_MAXUINT,
+      0,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_ID, param_spec);
 
   param_spec = g_param_spec_uint ("media-type", "Stream media type",
-                                  "A constant indicating which media type the "
-                                  "stream carries.",
-                                  TP_MEDIA_STREAM_TYPE_AUDIO,
-                                  TP_MEDIA_STREAM_TYPE_VIDEO,
-                                  TP_MEDIA_STREAM_TYPE_AUDIO,
-                                  G_PARAM_CONSTRUCT_ONLY |
-                                  G_PARAM_READWRITE |
-                                  G_PARAM_STATIC_NAME |
-                                  G_PARAM_STATIC_BLURB);
+      "A constant indicating which media type the stream carries.",
+      TP_MEDIA_STREAM_TYPE_AUDIO, TP_MEDIA_STREAM_TYPE_VIDEO,
+      TP_MEDIA_STREAM_TYPE_AUDIO,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_MEDIA_TYPE, param_spec);
 
   param_spec = g_param_spec_uint ("state", "Connection state",
-                                  "Connection state of the media stream",
-                                  TP_MEDIA_STREAM_STATE_DISCONNECTED,
-                                  TP_MEDIA_STREAM_STATE_CONNECTED,
-                                  TP_MEDIA_STREAM_STATE_DISCONNECTED,
-                                  G_PARAM_CONSTRUCT |
-                                  G_PARAM_READWRITE |
-                                  G_PARAM_STATIC_NAME |
-                                  G_PARAM_STATIC_BLURB);
-  g_object_class_install_property (object_class,
-                                   PROP_STATE,
-                                   param_spec);
+      "Connection state of the media stream",
+      TP_MEDIA_STREAM_STATE_DISCONNECTED, TP_MEDIA_STREAM_STATE_CONNECTED,
+      TP_MEDIA_STREAM_STATE_DISCONNECTED,
+      G_PARAM_CONSTRUCT | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
+  g_object_class_install_property (object_class, PROP_STATE, param_spec);
 
   /* We don't change the following two as individual properties
    * after construction, use tpsip_media_stream_set_direction() */
 
   param_spec = g_param_spec_uint ("direction", "Stream direction",
-                                  "A value indicating the current "
-                                        "direction of the stream",
-                                  TP_MEDIA_STREAM_DIRECTION_NONE,
-                                  TP_MEDIA_STREAM_DIRECTION_BIDIRECTIONAL,
-                                  TP_MEDIA_STREAM_DIRECTION_BIDIRECTIONAL,
-                                  G_PARAM_CONSTRUCT_ONLY |
-                                  G_PARAM_READWRITE |
-                                  G_PARAM_STATIC_NAME |
-                                  G_PARAM_STATIC_BLURB);
+      "A value indicating the current direction of the stream",
+      TP_MEDIA_STREAM_DIRECTION_NONE, TP_MEDIA_STREAM_DIRECTION_BIDIRECTIONAL,
+      TP_MEDIA_STREAM_DIRECTION_BIDIRECTIONAL,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class, PROP_DIRECTION, param_spec);
 
   param_spec = g_param_spec_uint ("pending-send-flags", "Pending send flags",
-                                  "Flags indicating the current "
-                                        "pending send state of the stream",
-                                  0,
-                                  TP_MEDIA_STREAM_PENDING_LOCAL_SEND
-                                        | TP_MEDIA_STREAM_PENDING_REMOTE_SEND,
-                                  0,
-                                  G_PARAM_CONSTRUCT_ONLY |
-                                  G_PARAM_READWRITE |
-                                  G_PARAM_STATIC_NAME |
-                                  G_PARAM_STATIC_BLURB);
+      "Flags indicating the current pending send state of the stream",
+      0,
+      TP_MEDIA_STREAM_PENDING_LOCAL_SEND | TP_MEDIA_STREAM_PENDING_REMOTE_SEND,
+      0,
+      G_PARAM_CONSTRUCT_ONLY | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class,
                                    PROP_PENDING_SEND_FLAGS,
                                    param_spec);
 
   param_spec = g_param_spec_boolean ("hold-state", "Hold state",
-                                     "Hold state of the media stream "
-                                        "as reported by the stream engine",
-                                     FALSE,
-                                     G_PARAM_CONSTRUCT |
-                                     G_PARAM_READWRITE |
-                                     G_PARAM_STATIC_NAME |
-                                     G_PARAM_STATIC_BLURB);
+      "Hold state of the media stream as reported by the stream engine",
+      FALSE,
+      G_PARAM_CONSTRUCT | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS);
   g_object_class_install_property (object_class,
                                    PROP_HOLD_STATE,
                                    param_spec);
@@ -450,6 +416,21 @@ tpsip_media_stream_class_init (TpsipMediaStreamClass *klass)
                   NULL, NULL,
                   g_cclosure_marshal_VOID__VOID,
                   G_TYPE_NONE, 0);
+
+/* Regexps for the name and the value parts of the parameter syntax */
+#define FMTP_TOKEN_ATTR "([-A-Za-z0-9_.]+)"
+#define FMTP_TOKEN_VALUE "([^;\"\\s]+|\"([^\"\\\\]|\\\\.)*\")"
+/* Indexes of the respective match groups in the whole regexp below */
+#define FMTP_MATCH_ATTR 1
+#define FMTP_MATCH_VALUE 2
+
+  fmtp_attr_regex = g_regex_new (FMTP_TOKEN_ATTR
+                                   "\\s*=\\s*"
+                                   FMTP_TOKEN_VALUE,
+                                 G_REGEX_RAW,
+                                 0 /* G_REGEX_MATCH_ANCHORED */,
+                                 NULL);
+  g_assert (fmtp_attr_regex != NULL);
 }
 
 void
@@ -906,6 +887,19 @@ tpsip_tp_stream_direction_from_remote (sdp_mode_t mode)
        | ((mode & sdp_sendonly)? TP_MEDIA_STREAM_DIRECTION_RECEIVE : 0);
 }
 
+static gboolean
+tpsip_sdp_codecs_differ (const sdp_rtpmap_t *m1, const sdp_rtpmap_t *m2)
+{
+  while (m1 != NULL && m2 != NULL)
+    {
+      if (sdp_rtpmap_cmp (m1, m2) != 0)
+        return TRUE;
+      m1 = m1->rm_next;
+      m2 = m2->rm_next;
+    }
+  return m1 != NULL || m2 != NULL;
+}
+
 /*
  * Returns stream direction as requested by the latest local or remote
  * direction change.
@@ -938,13 +932,15 @@ priv_get_requested_direction (TpsipMediaStreamPrivate *priv)
  */
 gboolean
 tpsip_media_stream_set_remote_media (TpsipMediaStream *stream,
-                                   const sdp_media_t *new_media,
-                                   guint direction_up_mask)
+                                     const sdp_media_t *new_media,
+                                     guint direction_up_mask,
+                                     guint pending_send_mask)
 {
   TpsipMediaStreamPrivate *priv;
   sdp_connection_t *sdp_conn;
   const sdp_media_t *old_media;
   gboolean transport_changed = TRUE;
+  gboolean codecs_changed = TRUE;
   guint old_direction;
   guint new_direction;
 
@@ -954,7 +950,7 @@ tpsip_media_stream_set_remote_media (TpsipMediaStream *stream,
 
   /* Do sanity checks */
 
-  g_return_val_if_fail (new_media != NULL, -1);
+  g_return_val_if_fail (new_media != NULL, FALSE);
 
   if (new_media->m_rejected || new_media->m_port == 0)
     {
@@ -975,6 +971,12 @@ tpsip_media_stream_set_remote_media (TpsipMediaStream *stream,
       return FALSE;
     }
 
+  if (new_media->m_rtpmaps == NULL)
+    {
+      g_warning ("Stream %u: no remote codecs", priv->id);
+      return FALSE;
+    }
+
   /* Note: always update the pointer to the current media structure
    * because of memory management done in the session object */
   old_media = priv->remote_media;
@@ -995,12 +997,16 @@ tpsip_media_stream_set_remote_media (TpsipMediaStream *stream,
    * if it's allowed to */
   new_direction &= old_direction | direction_up_mask;
 
-  /* Check if the transport candidate needs to be changed */
   if (old_media != NULL)
     {
+      /* Check if the transport candidate needs to be changed */
       if (!sdp_connection_cmp (sdp_media_connections (old_media), sdp_conn))
         transport_changed = FALSE;
 
+      /* Check if the codec list needs to be updated */
+      codecs_changed = tpsip_sdp_codecs_differ (old_media->m_rtpmaps,
+                                                new_media->m_rtpmaps);
+
       /* Disable sending at this point if it will be disabled
        * accordingly to the new direction */
       priv_update_sending (stream,
@@ -1014,21 +1020,25 @@ tpsip_media_stream_set_remote_media (TpsipMediaStream *stream,
 
   if (transport_changed)
     {
-     /* Make sure we stop sending before we use the new set of codecs
-      * intended for the new connection */
-      tpsip_media_stream_set_sending (stream, FALSE);
+      /* Make sure we stop sending before we use the new set of codecs
+       * intended for the new connection */
+      if (codecs_changed) 
+        tpsip_media_stream_set_sending (stream, FALSE);
 
       push_remote_candidates (stream);
     }
 
-  if (!priv->codec_intersect_pending)
+  if (codecs_changed)
     {
-      priv->codec_intersect_pending = TRUE;
-      push_remote_codecs (stream);
-    }
-  else
-    {
-      priv->push_remote_codecs_pending = TRUE;
+      if (!priv->codec_intersect_pending)
+        {
+          priv->codec_intersect_pending = TRUE;
+          push_remote_codecs (stream);
+        }
+      else
+        {
+          priv->push_remote_codecs_pending = TRUE;
+        }
     }
 
   /* TODO: this will go to session change commit code */
@@ -1038,10 +1048,10 @@ tpsip_media_stream_set_remote_media (TpsipMediaStream *stream,
    *       (and playing status?) once remote info set */
   push_active_candidate_pair (stream);
 
-  /* Set the final direction and clear remote pending send flag */
+  /* Set the final direction and update pending send flags */
   tpsip_media_stream_set_direction (stream,
                                     new_direction,
-                                    TP_MEDIA_STREAM_PENDING_LOCAL_SEND);
+                                    pending_send_mask);
 
   return TRUE;
 }
@@ -1061,8 +1071,6 @@ tpsip_tp_media_type (sdp_media_e sip_mtype)
       case sdp_media_video: return TP_MEDIA_STREAM_TYPE_VIDEO; 
       default: return G_MAXUINT;
     }
-
-  g_assert_not_reached();
 }
 
 /**
@@ -1266,61 +1274,114 @@ priv_generate_sdp (TpsipMediaStream *self)
   g_signal_emit (self, signals[SIG_READY], 0);
 }
 
-static void
-priv_parse_fmtp (su_home_t *home, const char *fmtp, GHashTable *param_hash)
+static gchar *
+priv_fmtp_fetch_value (const GMatchInfo *match)
 {
-  const msg_param_t *param_list = NULL;
-  char *s;
-  const char *avpair;
-  const char *pv;
-  char *attr;
-  int i;
+  const gchar *str;
+  gchar *value;
+  gchar *pv;
+  gint vstart = 0;
+  gint vend = 0;
+  gint i;
 
-  if (!fmtp)
+  str = g_match_info_get_string (match);
+  g_match_info_fetch_pos (match, FMTP_MATCH_VALUE, &vstart, &vend);
+
+  if (vstart < vend - 1 && str[vstart] == '"' && str[vend - 1] == '"')
+    {
+      /* Unescape the quoted pair contents */
+
+      /* The buffer should be enough for the non-collapsed contents
+       * minus the quotes */
+      value = g_malloc ((vend - 1 - vstart) * sizeof (gchar));
+      pv = value;
+      for (i = vstart + 1; i < vend - 1; i++)
+        {
+          if (str[i] == '\\')
+            ++i;
+          *pv++ = str[i];
+        }
+      *pv = '\0';
+    }
+  else
+    {
+      /* Pass as is */
+      value = g_strndup (str + vstart, vend - vstart);
+    }
+
+  return value;
+}
+
+static int
+priv_fmtp_match_length (const GMatchInfo *match)
+{
+  gint attr_start = 0;
+  gint attr_end = 0;
+
+  g_match_info_fetch_pos (match, 0, &attr_start, &attr_end);
+
+  g_assert (attr_start == 0);
+  g_assert (attr_end > 0);
+
+  return attr_end;
+}
+
+static void
+priv_fmtp_parse (const char *fmtp, GHashTable *param_hash)
+{
+  GMatchInfo *match;
+  gchar *name;
+  gchar *value;
+  const char *p = fmtp;
+
+  if (fmtp == NULL)
     return;
 
-  s = su_strdup (home, fmtp);
-  if (msg_avlist_d (home, &s, &param_list) < 0)
+  for (;;)
     {
-      DEBUG("format-specific parameters can't be parsed as an attribute list: %s", fmtp);
-      goto pass_as_is;
-    }
-  if (*s)
-    {
-      DEBUG("unexpected parser exit (whitespace?) while parsing format-specific parameters: %s", fmtp);
-      goto pass_as_is;
-    }
+      while (g_ascii_isspace (*p))
+        ++p;
 
-  g_assert (param_list != NULL);
+      if (*p == '\0')
+        break;
 
-  if (param_list[0] != NULL
-      && param_list[1] == NULL
-      && !strchr (param_list[0], '='))
-    {
-      DEBUG("passing freeform parameter string verbatim: %s", fmtp);
-      goto pass_as_is;
-    }
+      g_regex_match (fmtp_attr_regex, p, G_REGEX_MATCH_ANCHORED, &match);
+      if (!g_match_info_matches (match))
+        goto pass_as_is;
 
-  for (i = 0; (avpair = param_list[i]) != NULL; i++)
-    {
-      pv = strchr (avpair, '=');
-      if (pv)
+      name = g_match_info_fetch (match, FMTP_MATCH_ATTR);
+
+      value = priv_fmtp_fetch_value (match);
+
+      p += priv_fmtp_match_length (match);
+
+      g_match_info_free (match);
+
+      DEBUG("parsed parameter %s=%s", name, value);
+
+      g_hash_table_insert (param_hash, name, value);
+
+      while (g_ascii_isspace (*p))
+        ++p;
+
+      switch (*p)
         {
-          attr = su_strndup (home, avpair, pv - avpair);
-          DEBUG("parsed attribute-value pair %s=%s", attr, pv + 1);
-          g_hash_table_insert (param_hash, attr, (gpointer) (pv + 1));
-        }
-      else
-        {
-          DEBUG("parsed freeform parameter %s", avpair);
-          g_hash_table_insert (param_hash, (gpointer) avpair, NULL);
+        case '\0':
+          break;
+        case ';':
+          ++p;
+          break;
+        default:
+          goto pass_as_is;
         }
     }
 
   return;
 
 pass_as_is:
-  g_hash_table_insert (param_hash, "", (gpointer) fmtp);
+  DEBUG("failed to parse format-specific parameters as an attribute-value list: %s", fmtp);
+  g_hash_table_remove_all (param_hash);
+  g_hash_table_insert (param_hash, g_strdup (""), g_strdup (fmtp));
 }
 
 /**
@@ -1332,11 +1393,11 @@ static void push_remote_codecs (TpsipMediaStream *stream)
 {
   TpsipMediaStreamPrivate *priv;
   GPtrArray *codecs;
+  GHashTable *opt_params;
   GType codecs_type;
   GType codec_type;
   const sdp_media_t *sdpmedia;
   const sdp_rtpmap_t *rtpmap;
-  su_home_t temphome[1] = { SU_HOME_INIT(temphome) };
 
   DEBUG ("enter");
 
@@ -1360,21 +1421,21 @@ static void push_remote_codecs (TpsipMediaStream *stream)
   codecs_type = TP_ARRAY_TYPE_MEDIA_STREAM_HANDLER_CODEC_LIST;
 
   codecs = dbus_g_type_specialized_construct (codecs_type);
+  opt_params = g_hash_table_new_full (g_str_hash,
+                                      g_str_equal,
+                                      g_free,
+                                      g_free);
 
   rtpmap = sdpmedia->m_rtpmaps;
   while (rtpmap)
     {
       GValue codec = { 0, };
-      GHashTable *opt_params;
 
       g_value_init (&codec, codec_type);
       g_value_take_boxed (&codec,
                           dbus_g_type_specialized_construct (codec_type));
 
-      /* Note: all strings in opt_params hash are either allocated at temphome
-       * or referenced as is from the rtpmap structure */
-      opt_params = g_hash_table_new_full (g_str_hash, g_str_equal, NULL, NULL);
-      priv_parse_fmtp (temphome, rtpmap->rm_fmtp, opt_params);
+      priv_fmtp_parse (rtpmap->rm_fmtp, opt_params);
 
       /* RFC2327: see "m=" line definition 
        *  - note, 'encoding_params' is assumed to be channel
@@ -1395,22 +1456,20 @@ static void push_remote_codecs (TpsipMediaStream *stream)
                               5, opt_params,
                               G_MAXUINT);
 
-      g_ptr_array_add (codecs, g_value_get_boxed (&codec));
+      g_hash_table_remove_all (opt_params);
 
-      g_hash_table_destroy (opt_params);
+      g_ptr_array_add (codecs, g_value_get_boxed (&codec));
 
       rtpmap = rtpmap->rm_next;
     }
-  
-  su_home_deinit (temphome);
+
+  g_hash_table_destroy (opt_params);
 
   SESSION_DEBUG(priv->session, "passing %d remote codecs to stream engine",
                 codecs->len);
 
-  if (codecs->len > 0) {
-    tp_svc_media_stream_handler_emit_set_remote_codecs (
+  tp_svc_media_stream_handler_emit_set_remote_codecs (
         (TpSvcMediaStreamHandler *)stream, codecs);
-  }
 
   g_boxed_free (codecs_type, codecs);
 }
@@ -1420,6 +1479,7 @@ static void push_remote_candidates (TpsipMediaStream *stream)
   TpsipMediaStreamPrivate *priv;
   GValue candidate = { 0 };
   GValue transport = { 0 };
+  GValue transport_rtcp = { 0 };
   GPtrArray *candidates;
   GPtrArray *transports;
   GType candidate_type;
@@ -1456,12 +1516,15 @@ static void push_remote_candidates (TpsipMediaStream *stream)
 
   port = (guint) media->m_port;
 
+  transports_type = TP_ARRAY_TYPE_MEDIA_STREAM_HANDLER_TRANSPORT_LIST;
+  transports = dbus_g_type_specialized_construct (transports_type);
+
   transport_type = TP_STRUCT_TYPE_MEDIA_STREAM_HANDLER_TRANSPORT;
   g_value_init (&transport, transport_type);
   g_value_take_boxed (&transport,
                       dbus_g_type_specialized_construct (transport_type));
   dbus_g_type_struct_set (&transport,
-                          0, 0,         /* component number */
+                          0, 1,         /* component number */
                           1, sdp_conn->c_address,
                           2, port,
                           3, TP_MEDIA_STREAM_BASE_PROTO_UDP,
@@ -1473,12 +1536,61 @@ static void push_remote_candidates (TpsipMediaStream *stream)
                           /* 9, "", */
                           G_MAXUINT);
 
-  DEBUG("remote address=<%s>, port=<%u>", sdp_conn->c_address, port);
-
-  transports_type = TP_ARRAY_TYPE_MEDIA_STREAM_HANDLER_TRANSPORT_LIST;
-  transports = dbus_g_type_specialized_construct (transports_type);
+  DEBUG("remote RTP address=<%s>, port=<%u>", sdp_conn->c_address, port);
   g_ptr_array_add (transports, g_value_get_boxed (&transport));
 
+  if (!tpsip_sdp_rtcp_bandwidth_throttled (media->m_bandwidths))
+    {
+      gboolean session_rtcp_enabled = TRUE;
+      g_object_get (priv->session,
+                    "rtcp-enabled", &session_rtcp_enabled,
+                    NULL);
+      if (session_rtcp_enabled)
+        {
+          const sdp_attribute_t *rtcp_attr;
+          const char *rtcp_address;
+          guint rtcp_port;
+
+          /* Get the port and optional address for RTCP accordingly to RFC 3605 */
+          rtcp_address = sdp_conn->c_address;
+          rtcp_attr = sdp_attribute_find (media->m_attributes, "rtcp");
+          if (rtcp_attr == NULL || rtcp_attr->a_value == NULL)
+            {
+              rtcp_port = port + 1;
+            }
+          else
+            {
+              const char *rest;
+              rtcp_port = (guint) g_ascii_strtoull (rtcp_attr->a_value,
+                                                    (gchar **) &rest,
+                                                    10);
+              if (rtcp_port != 0
+                  && (strncmp (rest, " IN IP4 ", 8) == 0
+                      || strncmp (rest, " IN IP6 ", 8) == 0))
+                rtcp_address = rest + 8;
+            }
+
+          g_value_init (&transport_rtcp, transport_type);
+          g_value_take_boxed (&transport_rtcp,
+                              dbus_g_type_specialized_construct (transport_type));
+          dbus_g_type_struct_set (&transport_rtcp,
+                                  0, 2,         /* component number */
+                                  1, rtcp_address,
+                                  2, rtcp_port,
+                                  3, TP_MEDIA_STREAM_BASE_PROTO_UDP,
+                                  4, "RTCP",
+                                  5, "AVP",
+                                  /* 6, 0.0f, */
+                                  7, TP_MEDIA_STREAM_TRANSPORT_TYPE_LOCAL,
+                                  /* 8, "", */
+                                  /* 9, "", */
+                                  G_MAXUINT);
+
+          DEBUG("remote RTCP address=<%s>, port=<%u>", rtcp_address, rtcp_port);
+          g_ptr_array_add (transports, g_value_get_boxed (&transport_rtcp));
+        }
+    }
+
   g_free (priv->remote_candidate_id);
   candidate_id = g_strdup_printf ("L%u", ++priv->remote_candidate_counter);
   priv->remote_candidate_id = candidate_id;
@@ -1557,6 +1669,67 @@ priv_marshal_param (gpointer key,
     g_string_append_printf (params, "%s=%s", name, value);
 }
 
+static void
+priv_append_rtpmaps (const GPtrArray *codecs, GString *mline, GString *alines)
+{
+  GValue codec = { 0, };
+  gchar *co_name = NULL;
+  guint co_id;
+  /* guint co_type; */
+  guint co_clockrate;
+  guint co_channels;
+  GHashTable *co_params = NULL;
+  guint i;
+
+  g_value_init (&codec, TP_STRUCT_TYPE_MEDIA_STREAM_HANDLER_CODEC);
+
+  for (i = 0; i < codecs->len; i++)
+    {
+      g_value_set_static_boxed (&codec, g_ptr_array_index (codecs, i));
+
+      dbus_g_type_struct_get (&codec,
+                              0, &co_id,
+                              1, &co_name,
+                              /* 2, &co_type, */
+                              3, &co_clockrate,
+                              4, &co_channels,
+                              5, &co_params,
+                              G_MAXUINT);
+
+      /* g_return_if_fail (co_type == priv->media_type); */
+
+      /* Add rtpmap entry to the a= lines */
+      g_string_append_printf (alines,
+                              "a=rtpmap:%u %s/%u",
+                              co_id,
+                              co_name,
+                              co_clockrate);
+      if (co_channels > 1)
+        g_string_append_printf (alines, "/%u", co_channels);
+      g_string_append (alines, "\r\n");
+
+      /* Marshal parameters into the fmtp attribute */
+      if (g_hash_table_size (co_params) != 0)
+        {
+          GString *fmtp_value;
+          g_string_append_printf (alines, "a=fmtp:%u ", co_id);
+          fmtp_value = g_string_new (NULL);
+          g_hash_table_foreach (co_params, priv_marshal_param, fmtp_value);
+          g_string_append (alines, fmtp_value->str);
+          g_string_free (fmtp_value, TRUE);
+          g_string_append (alines, "\r\n");
+        }
+
+      /* Add PT id to the m= line */
+      g_string_append_printf (mline, " %u", co_id);
+
+      g_free (co_name);
+      co_name = NULL;
+      g_hash_table_destroy (co_params);
+      co_params = NULL;
+    }
+}
+
 /**
 * Refreshes the local SDP based on Farsight stream, and current
 * object, state.
@@ -1569,34 +1742,26 @@ priv_update_local_sdp(TpsipMediaStream *stream)
   GString *alines;
   gchar *cline;
   GValue transport = { 0 };
-  GValue codec = { 0, };
-  const GPtrArray *codecs, *candidates;
+  const GPtrArray *candidates;
   gchar *tr_addr = NULL;
   /* gchar *tr_user = NULL; */
   /* gchar *tr_pass = NULL; */
   gchar *tr_subtype = NULL;
   gchar *tr_profile = NULL;
   guint tr_port;
-  /* guint tr_component; */
-  guint tr_proto;
+  guint tr_component;
   /* guint tr_type; */
   /* gdouble tr_pref; */
+  guint rtcp_port = 0;
+  gchar *rtcp_address = NULL;
   const gchar *dirline;
-  const gchar *iproto;
   int i;
 
-  /* Note: impl limits ...
-   * - no multi-stream support
-   * - no IPv6 support (missing from the Farsight API?)
-   */
-
   priv = TPSIP_MEDIA_STREAM_GET_PRIVATE (stream);
 
   candidates = g_value_get_boxed (&priv->native_candidates);
-  codecs = g_value_get_boxed (&priv->native_codecs);
 
   g_value_init (&transport, TP_STRUCT_TYPE_MEDIA_STREAM_HANDLER_TRANSPORT);
-  g_value_init (&codec, TP_STRUCT_TYPE_MEDIA_STREAM_HANDLER_CODEC);
 
   /* Find the preferred candidate, if defined,
    * else the last acceptable candidate */
@@ -1606,28 +1771,50 @@ priv_update_local_sdp(TpsipMediaStream *stream)
       GValueArray *candidate;
       const gchar *candidate_id;
       const GPtrArray *ca_tports;
+      guint tr_proto = TP_MEDIA_STREAM_BASE_PROTO_UDP;
+      guint j;
 
       candidate = g_ptr_array_index (candidates, i);
       candidate_id =
                 g_value_get_string (g_value_array_get_nth (candidate, 0));
       ca_tports = g_value_get_boxed (g_value_array_get_nth (candidate, 1));
 
-      g_return_if_fail (ca_tports->len >= 1);
+      if (ca_tports->len == 0)
+        {
+          g_warning ("candidate '%s' lists no transports, skipping", candidate_id);
+          continue;
+        }
 
-      g_value_set_static_boxed (&transport, g_ptr_array_index (ca_tports, 0));
-
-      dbus_g_type_struct_get (&transport,
-                              /* 0, &tr_component, */
-                              1, &tr_addr,
-                              2, &tr_port,
-                              3, &tr_proto,
-                              4, &tr_subtype,
-                              5, &tr_profile,
-                              /* 6, &tr_pref, */
-                              /* 7, &tr_type, */
-                              /* 8, &tr_user, */
-                              /* 9, &tr_pass, */
-                              G_MAXUINT);
+      for (j = 0; j < ca_tports->len; j++)
+        {
+          g_value_set_static_boxed (&transport,
+                                    g_ptr_array_index (ca_tports, j));
+          dbus_g_type_struct_get (&transport,
+                                  0, &tr_component,
+                                  G_MAXUINT);
+          switch (tr_component)
+            {
+            case 1:     /* RTP */
+              dbus_g_type_struct_get (&transport,
+                                      1, &tr_addr,
+                                      2, &tr_port,
+                                      3, &tr_proto,
+                                      4, &tr_subtype,
+                                      5, &tr_profile,
+                                      /* 6, &tr_pref, */
+                                      /* 7, &tr_type, */
+                                      /* 8, &tr_user, */
+                                      /* 9, &tr_pass, */
+                                      G_MAXUINT);
+              break;
+            case 2:     /* RTCP */
+              dbus_g_type_struct_get (&transport,
+                                      1, &rtcp_address,
+                                      2, &rtcp_port,
+                                      G_MAXUINT);
+              break;
+            }
+        }
 
       if (priv->native_candidate_id != NULL)
         {
@@ -1654,8 +1841,9 @@ priv_update_local_sdp(TpsipMediaStream *stream)
                           tr_subtype,
                           tr_profile);
 
-  iproto = (strchr (tr_addr, ':') == NULL)? "IP4" : "IP6";
-  cline = g_strdup_printf ("c=IN %s %s\r\n", iproto, tr_addr);
+  cline = g_strdup_printf ("c=IN %s %s\r\n",
+                           (strchr (tr_addr, ':') == NULL)? "IP4" : "IP6",
+                           tr_addr);
 
   switch (priv_get_requested_direction (priv))
     {
@@ -1677,59 +1865,32 @@ priv_update_local_sdp(TpsipMediaStream *stream)
 
   alines = g_string_new (dirline);
 
-  for (i = 0; i < codecs->len; i++) {
-    guint co_id, co_type, co_clockrate, co_channels;
-    gchar *co_name;
-    GHashTable *co_params;
+  if (rtcp_address != NULL)
+    {
+      /* Add RTCP attribute as per RFC 3605 */
+      if (strcmp (rtcp_address, tr_addr) != 0)
+        {
+          g_string_append_printf (alines,
+                                  "a=rtcp:%u IN %s %s\r\n",
+                                  rtcp_port,
+                                  (strchr (rtcp_address, ':') == NULL)
+                                        ? "IP4" : "IP6",
+                                  rtcp_address);
+        }
+      else if (rtcp_port != tr_port + 1)
+        {
+          g_string_append_printf (alines,
+                                  "a=rtcp:%u\r\n",
+                                  rtcp_port);
+        }
+    }
 
-    g_value_set_static_boxed (&codec, g_ptr_array_index (codecs, i));
-
-    dbus_g_type_struct_get (&codec,
-			    0, &co_id,
-			    1, &co_name,
-			    2, &co_type, 
-			    3, &co_clockrate,
-			    4, &co_channels,
-                            5, &co_params,
-			    G_MAXUINT);
-
-    g_return_if_fail (co_type == priv->media_type);
-
-    /* Add rtpmap entry to media a-lines */
-    g_string_append_printf (alines,
-                            "a=rtpmap:%u %s/%u",
-                            co_id,
-                            co_name,
-                            co_clockrate);
-    if (co_channels > 1)
-      g_string_append_printf (alines, "/%u", co_channels);
-    g_string_append (alines, "\r\n");
-
-    /* Marshal parameters into the fmtp attribute */
-    if (g_hash_table_size (co_params) != 0)
-      {
-        GString *fmtp_value;
-        g_string_append_printf (alines, "a=fmtp:%u ", co_id);
-        fmtp_value = g_string_new (NULL);
-        g_hash_table_foreach (co_params, priv_marshal_param, fmtp_value);
-        g_string_append (alines, fmtp_value->str);
-        g_string_free (fmtp_value, TRUE);
-        g_string_append (alines, "\r\n");
-      }
-
-    /* Add PT id to mline */
-    g_string_append_printf (mline, " %u", co_id);
-
-    g_free (co_name);
-    g_hash_table_destroy (co_params);
-  }
+  priv_append_rtpmaps (g_value_get_boxed (&priv->native_codecs),
+                       mline, alines);
 
   g_free(priv->stream_sdp);
   priv->stream_sdp = g_strconcat(mline->str, "\r\n",
                                  cline,
-#if 0   /* Some Cisco GWs do not like these */
-                                 "b=RS:0\r\nb=RR:0\r\n",
-#endif
                                  alines->str,
                                  NULL);
 
@@ -1738,6 +1899,7 @@ priv_update_local_sdp(TpsipMediaStream *stream)
   g_free (tr_subtype);
   /* g_free (tr_user); */
   /* g_free (tr_pass); */
+  g_free (rtcp_address);
 
   g_string_free (mline, TRUE);
   g_free (cline);
