@@ -1,25 +1,26 @@
 /*
- * sip-connection.c - Source for SIPConnection
+ * sip-connection.c - Source for TpsipConnection
  * Copyright (C) 2005 Collabora Ltd.
- * Copyright (C) 2005,2006 Nokia Corporation
+ * Copyright (C) 2005-2008 Nokia Corporation
  *   @author Kai Vehmanen <first.surname@nokia.com>
  *   @author Martti Mela <first.surname@nokia.com>
- *   @author Mikhail Zabaluev <first.surname@nokia.com>
+ *   @author Mikhail Zabaluev <mikhail.zabaluev@nokia.com>
  *
  * Based on telepathy-gabble implementation (gabble-connection).
  *   @author See gabble.c
  *
- * This library is free software; you can redistribute it and/or
- * modify it under the terms of the GNU Lesser General Public License
- * version 2.1 as published by the Free Software Foundation.
+ * This work is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU Lesser General Public
+ * License as published by the Free Software Foundation; either
+ * version 2.1 of the License, or (at your option) any later version.
  *
- * This library is distributed in the hope that it will be useful,
+ * This work is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
  * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the Free Software
+ * License along with this work; if not, write to the Free Software
  * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
@@ -44,7 +45,7 @@ static TpBaseConnectionManager *
 construct_cm (void)
 {
   return (TpBaseConnectionManager *)g_object_new (
-      SIP_TYPE_CONNECTION_MANAGER, NULL);
+      TPSIP_TYPE_CONNECTION_MANAGER, NULL);
 }
 
 #ifdef ENABLE_SYSLOG
@@ -138,14 +139,16 @@ main (int argc, char** argv)
   gpointer logdata;
 
 #ifdef ENABLE_DEBUG
-  sip_debug_set_flags_from_env ();
+  tpsip_debug_set_flags_from_env ();
 
-  if (g_getenv ("SOFIASIP_PERSIST"))
+  if (g_getenv ("TPSIP_PERSIST") || g_getenv ("SOFIASIP_PERSIST"))
     {
       tp_debug_set_persistent (TRUE);
     }
 #endif
 
+  tp_debug_divert_messages (g_getenv ("TPSIP_LOGFILE"));
+
   logdata = sofia_log_init ();
 
   status = tp_run_connection_manager ("telepathy-sofiasip", VERSION,
